{-----------------------------------------------------------------------------}
{-                                                                           -}
{-   Projet : RPNCALC                                                        -}
{-   Module : RPNENGINE -- the actual data stack, with all math functions.   -}
{-                                                                           -}
{-   Version : 0.9 (September 11, 1995)                                      -}
{-                                                                           -}
{-   Written by Cyrille Chplov <100420.1574@compuserve.com>                -}
{-   Graphics by Pierre Chplov                                             -}
{-                                                                           -}
{-  These source files and the associated executable are freeware.           -}
{-  Constructive feedback is greatly encouraged !                            -}
{-                                                                           -}
{-  Warranty : none. Use at your own risk.                                   -}
{-                                                                           -}
{-----------------------------------------------------------------------------}

unit Rpnengine;
interface
uses Sysutils,Classes,Dialogs;

Type
    Float = Double;
    PFloat = ^Float;

Type
    TDegMode = (dmDegrees,dmGrades,dmRadians);

Type
    TRPNStack = Class (TList)
       private
         Procedure SetFloat(i : integer; f : Float);
         Function GetFloat(i : integer) : float;
       public
         Procedure Push(f : float);
         Function Pop : float;

         Procedure _Dropn(i : integer);

       public  { RPL Language primitives }

         Function CheckParamCount(const s : string; i : integer) : boolean;
         Procedure InvalidValue(const s : string);

         Procedure Insert(i : integer; f : float);
         Function Delete(i : integer) : float;

         Function DUP : boolean;
         Function ADD : boolean;
         Function SUB : boolean;
         Function MUL : boolean;
         Function fDIV : boolean;
         Function NEG : boolean;
         Function DROP : boolean;
         Function CLR : Boolean;

         Function SWAP : Boolean;

       public { fonctions scientifiques }
         degMode : TDegMode;

         Function PopRad : float;
         Procedure PushRad(f : Float);

         Function COS : boolean;
         Function SIN : boolean;
         Function TAN : boolean;

         Function CH : boolean;
         Function SH : boolean;
         Function TH : boolean;

         Function ArcCos : Boolean;
         Function ArcTan : Boolean;
         Function ArcSin : Boolean;

         Function ArgCosH : Boolean;
         Function ArgTanH : Boolean;
         Function ArgSinH : Boolean;

         Function SQ : Boolean;
         Function SQRT : Boolean;

         Function POWER : Boolean;
         Function XROOT : Boolean;

         Function EXP : Boolean;
         Function LN : Boolean;

         Function ALOG : Boolean;
         Function LOG : Boolean;

         Function FACT : Boolean;
         Function COMB : Boolean;
         Function PERM  : Boolean;


         Function _pi : Boolean;
       public
         property Items[i : integer] : float read GetFloat write SetFloat;
       end;

implementation

Procedure TRPNStack.SetFloat(i : integer; f : Float);
begin
     PFloat(inherited Items[Count-i])^ := f;
end;

Function TRPNStack.GetFloat(i : integer) : Float;
begin
     Result := PFloat(inherited items[count-i])^;
end;

Procedure TRPNStack.Push(f : float);
var
   pf : PFloat;
begin
     new(pf);
     pf^ := f;
     inherited Add(pf);
end;

Function TRPNStack.Pop : float;
var
   pf : PFloat;
begin
     pf := PFloat(inherited items[count-1]);
     inherited delete(count-1);
     result := pf^;
     dispose(pf);
end;

Procedure TRPNStack._Dropn(i : integer);
begin
     while i > 0 do
     begin
          pop; dec(i);
     end;
end;

Procedure TRPNStack.Insert(i : integer; f : float);
var
   pf : PFloat;
begin
     new(pf); pf^ := f;
     inherited Insert(Count-i,pf);
end;

Function TRPNStack.Delete(i : integer) : Float;
var
   pf : PFloat;
begin
     pf := PFloat(inherited Items[Count-i]);
     Result := pf^;
     Dispose(pf);
     inherited Delete(Count-i);
end;

Function TRPNStack.CheckParamCount(const s : string;i : integer) : boolean;
begin
     Result := Count >= i;
     if not Result then MessageDlg( s+Format(' Error : %d argument(s) expected',[i]),
                                    mtWarning,[mbOK],0);
end;

Procedure TRPNStack.InvalidValue(const s : string);
begin
     MessageDlg(s+' Error : invalid value',mtWarning,[mbOK],0);
end;

Function TRPNStack.DUP : boolean;
var
   f : Float;
begin
     Result := CheckParamCount('DUP',1);
     if not Result then exit;
     f := Pop;
     Push(f);
     Push(f);
end;



Function TRPNStack.ADD : boolean;
begin
     Result := CheckParamCount('+',2);
     if not Result then exit;
     Push(Pop + Pop);
end;

Function TRPNStack.SUB : boolean;
begin
     Result := CheckParamCount('-',2);
     if not Result then exit;
     Push( (-Pop) + Pop);
end;

Function TRPNStack.MUL : boolean;
begin
     Result := CheckParamCount('*',2);
     if not Result then exit;
     Push(Pop * Pop);
end;

Function TRPNStack.fDIV : boolean;
var
   f : float;
begin
     Result := CheckParamCount('/',2);
     if not Result then exit;
     f := pop;
     if f = 0 then
     begin
          InvalidValue('/');
          result := false;
          exit;
     end;

     Push((1/f) * Pop);
end;

Function TRPNStack.NEG : boolean;
begin
     Result := CheckParamCount('NEG',1);
     if not Result then exit;
     Push(-Pop);
end;

Function TRPNStack.DROP : boolean;
begin
     Result := CheckParamCount('DROP',1);
     if not Result then exit;
     _dropn(1);
end;

Function TRPNStack.CLR : Boolean;
begin
     Result := True;
     Clear;
end;

Function TRPNStack.SWAP : Boolean;
var
   f1,f2 : float;
begin
     Result := CheckParamCount('SWAP',2);
     if not Result then exit;
     f1 := Pop; f2 := pop;
     push(f1); push(f2);
end;

{
         Function COS : boolean;
         Function SIN : boolean;
         Function TAN : boolean;
         Function CH : boolean;
         Function SH : boolean;
         Function TH : boolean;
}

Function TRPNStack.PopRad;
begin
     Result := pop;
     case degMode of
       dmDegrees : Result := (Result * PI) / 180;
       dmGrades : Result :=  (Result * PI) / 200;
       { dmRadians : rien  faire }
     end;
end;

Procedure TRPNStack.PushRad;
begin
     case degMode of
       dmDegrees : f := (f * 180) / PI;
       dmGrades : f :=  (f * 200) / PI;
       { dmRadians : rien  faire }
     end;
     Push(f);
end;

Function TRPNStack.SIN;
begin
     Result := CheckParamCount('SIN',1);
     if not Result then exit;
     Push(System.Sin(PopRad));
end;

Function TRPNStack.COS;
begin
     Result := CheckParamCount('COS',1);
     if not Result then exit;
     Push(System.COS(PopRad));
end;

Function TRPNStack.TAN;
var
   f : Float;
begin
     Result := CheckParamCount('TAN',1);
     if not Result then exit;
     (* DUP; COS; SWAP; SIN; DIV;  { bon, l, c'tait pour rire... } *)
     f := PopRad;
     if system.Cos(f) = 0 then
     begin
          InvalidValue('TAN');
          result := false;
          exit;
     end;

     Push(System.SIN(f) / System.COS(f) );
end;


Function TRPNStack.ArcTan;
begin
     Result := CheckParamCount('ATAN',1);
     if not Result then exit;
     PushRad(System.Arctan(Pop));
end;

Function TRPNStack.ArcCos;
var
   y : Float;
begin
     Result := CheckParamCount('ACOS',1);
     if not Result then exit;

     { Pascal ne fournit pas de fonction ArcCos.
       Le FPU Intel devrait la fournir depuis le 386, mais j'envisage l'usage de
       cette applet dans une VDM de machine RISC NT (ou pourquoi pas, un PowerMac,
       maudit soit-il). Par ailleurs, je rpugne  utiliser ici de l'assembleur,
       car je vais me servir de cette applet pour tester Delphi32. Donc exit l'ASM.

       On va donc utiliser le fait que, pour x <> 0 [ PI prs], en posant t = tan(x/2),
       on ait : cos(x) = (1-t^2)/(1+t^2).    Posons y = cos(x).
             y = (1-t^2) / (1+t^2)
             y*(1+t^2) = 1-t^2
             t^2 * (1+y) + (y-1) = 0      D = 0^2 - 4*(y-1)*(y+1) = 4*(1-y^2)
             t+ = ( -0 + sqrt(D) ) / 2*(1+y)      (t- abandonn, par convention)

             t = sqrt(1-y^2)/(1+y) = (rappel) tan(x/2)    Gagn !

       finalement, on obtient donc :

             x = 2*ArcTan( sqrt(1-y^2) / (1+y) )

       que Saint Niklaus nous permet de compiler, mais seulement le 6 dcembre.
     }

     y := Pop;
     if abs(y) > 1 then
     begin
          InvalidValue('ACOS');
          result := false;
          exit;
     end;

     if y = -1 then     { cas particulier interdit par tan(x/2). L'autre cas (x=0) }
     begin              { est prolongeable par continuit dans la fonction x = g(y) }
          PushRad(PI);
          exit;
     end;

     PushRad( 2*System.Arctan( System.sqrt(1-(y*y)) / (1+y) ) );
end;

Function TRPNStack.ArcSin;
var
   y : Float;
begin
     Result := CheckParamCount('ASIN',1);
     if not Result then exit;

     { mme cirque pour la fonction ArcSin :
         y = sin(x) = 2*t / (1+t^2)      avec t = tan(x/2), x <> 0[PI]

         y(1+t^2) - 2*t = 0
         t^2 * y - 2*t + y = 0       Dr = 1 - y^2

         t+ = (- (1) + sqrt(Dr) ) / y     (t- abandonn)

         t = (sqrt(Dr) -1) / y

         x = - 2 * ArcTan( (sqrt(1-y^2)-1) / y )

     }

     y := Pop;

     if abs(y) > 1 then
     begin
          InvalidValue('ASIN');
          result := false;
          exit;
     end;

     if y = 0 then
     begin
          PushRad(0);
          exit;
     end;

     PushRad( -2 * System.ArcTan( (System.Sqrt(1-(y*y) )-1) / y ) );
end;

Function TRPNStack.CH;
var
   f : Float;
begin
     { Fonctions hyperboliques totalement absentes en Pascal.
       Fort heureusement, grce  l'ami Euler, ce n'est pas grave. }

     Result := CheckParamCount('COSH',1);
     if not Result then exit;
     f := Pop;

     Push (0.5 * (System.Exp(f) + System.Exp(-f)) );
end;

Function TRPNStack.SH;
var
   f : Float;
begin
     Result := CheckParamCount('SINH',1);
     if not Result then exit;
     f := Pop;

     Push (0.5 * (System.Exp(f) - System.Exp(-f)) );
end;

Function TRPNStack.TH;
var
   f,xf,xnf : Float;
begin
     Result := CheckParamCount('TANH',1);
     if not Result then exit;
     f := Pop;
     xf := System.exp(f); xnf := System.exp(-f);

     Push ( (xf - xnf) / (xf + xnf) );
end;

Function TRPNStack.ArgSinH;
var
   y : Float;
begin
     Result := CheckParamCount('ASINH',1);
     if not Result then exit;
     y := pop;

     push( system.ln( y + system.sqrt(y*y + 1) ) );
end;

Function TRPNStack.ArgCosH;
var
   y : Float;
begin
     Result := CheckParamCount('ACOSH',1);
     if not Result then exit;
     y := pop;
     if y < 0 then
     begin
          InvalidValue('ACOSH');
          result := false;
          exit;
     end;

     push( system.ln( y + system.sqrt(y*y - 1) ) );
end;

Function TRPNStack.ArgTanH;
var
   y : Float;
begin
     Result := CheckParamCount('ATANH',1);
     if not Result then exit;
     y := pop;
     if Abs(y) >= 1 then
     begin
          InvalidValue('ATANH');
          result := false;
          exit;
     end;

     push( 0.5*system.ln( (1+y)/(1-y) ) );
end;

Function TRPNStack.SQ;
begin
     Result := CheckParamCount('SQ',1);
     if not Result then exit;

     push(System.SQR(pop));
end;

Function TRPNStack.SQRT;
var
   x : float;
begin
     Result := CheckParamCount('SQRT',1);
     if not Result then exit;

     x := Pop;
     if x < 0 then
     begin
          InvalidValue('SQRT');
          Result := false;
          exit;
     end;


     push(system.sqrt(x));
end;

Function TRPNStack.POWER;
var
   y,x : Float;
   sgn : Float;
begin
     Result := CheckParamCount('POWER',2);
     if not Result then exit;

     x := pop;
     y := pop;

     { petite astuce : y < 0 est parfaitement lgal. Mais ln(y) ne passera pas ici,
       car les logs ngatifs = tabou (d'aprs l'honorable et inestimable Dan le Boucher)
       bon, par paresse de rechercher, on se contentera de l'erzatz : y/abs(y) }
     if y = 0 then Sgn := 0 else sgn := y / abs(y);

     Push( Sgn* system.exp( x* system.ln(abs(y)) ) );
end;

Function TRPNStack.XROOT;
var
   y,x : Float;
   sgn : Float;
begin
     Result := CheckParamCount('XROOT',2);
     if not Result then exit;

     x := pop;
     y := pop;

     if y <> 0 then sgn := y / abs(y);

     if (x = 0) or
        (y < 0) and ((Frac(x) <> 0) or (not Odd(Trunc(x))) )  then
     begin
          InvalidValue('XROOT');
          Result := False;
          exit;
     end;

     Push( Sgn* system.exp( (1/x) * system.ln(abs(y)) ) );
end;

var
   ln10 : float;

Function TRPNStack.EXP;
begin
     Result := CheckParamCount('EXP',1);
     if not Result then exit;

     push(System.EXP(pop));
end;

Function TRPNStack.ALOG;
begin
     Result := CheckParamCount('ALOG',1);
     if not Result then exit;

     push(System.EXP(pop * ln10));
end;

Function TRPNStack.LN;
var
   x : float;
begin
     Result := CheckParamCount('LN',1);
     if not Result then exit;

     x := Pop;
     if x <= 0 then
     begin
          InvalidValue('LN');
          Result := false;
          exit;
     end;

     push(system.ln(x));
end;

Function TRPNStack.LOG;
var
   x : float;
begin
     Result := CheckParamCount('LOG',1);
     if not Result then exit;

     x := Pop;
     if x <= 0 then
     begin
          InvalidValue('LOG');
          Result := false;
          exit;
     end;

     push(system.ln(x)/ln10);
end;

{
         Function FACT : Boolean;
         Function COMB : Boolean;
         Function ARR  : Boolean;
}

Function InternalFact(n : longint) : Float;
var
   i : longint;
begin
     Result := 1;
     for i := 1 to n do Result := Result * i;
end;


Function TRPNStack.FACT;
var
   x : float;
   n : longint;
   f : float;
begin
     Result := CheckParamCount('FACT',1);
     if not Result then exit;

     x := pop;
     n := trunc(x);
     if (n < 0) or (x <> n) then
     begin
          InvalidValue('FACT');
          result := false;
          exit;
     end;

     f := 1;
     try
        f := InternalFact(n);
     except
        InvalidValue('FACT');
        result := false;
        exit;
     end;

     push(f);
end;

Function TRPNStack.COMB;
var
   x,y : float;
   i,n,p : longint;
   f : float;
begin
     Result := CheckParamCount('COMB',1);
     if not Result then exit;

     x := pop; p := trunc(x);
     y := pop; n := trunc(y);
     if (n < 0) or (x <> p) or (y <> n) or (p < 0) or (p > n) then
     begin
          InvalidValue('COMB');
          result := false;
          exit;
     end;

     try
        f := InternalFact(n)/ (InternalFact(p)*InternalFact(n-p) );
     except
        InvalidValue('COMB');
        result := false;
        exit;
     end;

     push(f);
end;

Function TRPNStack.PERM;
var
   x,y : float;
   i,n,p : longint;
   f : float;
begin
     Result := CheckParamCount('PERM',1);
     if not Result then exit;

     x := pop; p := trunc(x);
     y := pop; n := trunc(y);
     if (n < 0) or (x <> p) or (y <> n) or (p < 0) or (p > n) then
     begin
          InvalidValue('PERM');
          result := false;
          exit;
     end;

     try
        f := InternalFact(n)/ (InternalFact(n-p) );
     except
        InvalidValue('PERM');
        result := false;
        exit;
     end;

     push(f);
end;



Function TRPNStack._pi;
begin
     Result := True;
     Push(pi);
end;

begin
     ln10 := ln(10);
end.

